/*****************************************************************************
 *
 * Copyright (C) 2018  Wilhelm Hagemeister <hm@igh.de>, with code from
 *                     Florian Pose <fp@igh.de>
 *
 * This file is part of the QtPdCom library.
 *
 * The QtPdCom library is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 *
 * The QtPdCom library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with the QtPdCom Library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 ****************************************************************************/

#include "ScalarVariant.h"
#include "ScalarVariant_p.h"

using QtPdCom::ScalarVariant;
using QtPdCom::ScalarVariantPrivate;

#include <QtPdCom1/Transmission.h>

#include <QStringList>

#include <QtDebug>

// #define PD_DEBUG_SCALARVARIANT

/****************************************************************************/

/** Constructor.
 */
ScalarVariant::ScalarVariant(QObject *parent):
    QObject(parent),
    d_ptr(new ScalarVariantPrivate(this))
{}


ScalarVariantPrivate::ScalarVariantPrivate(ScalarVariant *This):
    q_ptr(This),
    value(0),
    dataPresent(false)
{}

/****************************************************************************/

/** Destructor.
 */
ScalarVariant::~ScalarVariant() = default;

/****************************************************************************/

void ScalarVariantPrivate::updateConnection()
{
    Q_Q(ScalarVariant);

    if (process_disconnected) {
        q->disconnect(process_disconnected);
    }
    if (process_error) {
        q->disconnect(process_error);
    }


    if (!pd_connection.setVariable(*q)) {
        return;
    }
    process_disconnected = q->connect(
            pd_connection.getProcess(),
            &Process::disconnected,
            q,
            &ScalarVariant::clearData);
    process_error = q->connect(
            pd_connection.getProcess(),
            &Process::error,
            q,
            &ScalarVariant::clearData);
}

/****************************************************************************/

void QtPdCom::ScalarVariant::setConnection(const QtPdCom::PdConnection &conn)
{
    Q_D(ScalarVariant);
    d->pd_connection = conn;
    d->updateConnection();
}

/****************************************************************************/

void ScalarVariant::clearData()
{
    Q_D(ScalarVariant);

    d->value = 0;
    d->dataPresent = false;
    emit dataPresentChanged(d->dataPresent);
    emit valueChanged(d->value);
}

/****************************************************************************/

void ScalarVariant::setValue(QVariant value)
{
    Q_D(ScalarVariant);

    if (not d->dataPresent or getVariable().empty()) {
        return;
    }

    // nothing to write
    if (value == d->value) {
        return;
    }

    switch (getVariable().getTypeInfo().type) {
        case PdCom::TypeInfo::boolean_T:
        case PdCom::TypeInfo::uint8_T:
        case PdCom::TypeInfo::uint16_T:
        case PdCom::TypeInfo::uint32_T:
        case PdCom::TypeInfo::uint64_T:
            if (value.canConvert<uint64_t>()) {
                writeValue((uint64_t) value.toULongLong());
            }
            else {
                qWarning()
                        << "Variant datatype can't be converted to ULongLong";
            }
            break;
        case PdCom::TypeInfo::int8_T:
        case PdCom::TypeInfo::int16_T:
        case PdCom::TypeInfo::int32_T:
        case PdCom::TypeInfo::int64_T:
            if (value.canConvert<int64_t>()) {
                writeValue((int64_t) value.toLongLong());
            }
            else {
                qWarning()
                        << "Variant datatype can't be converted to LongLong";
            }
            break;
        case PdCom::TypeInfo::single_T:
        case PdCom::TypeInfo::double_T:
            if (value.canConvert<double>()) {
                writeValue(value.toDouble());
            }
            else {
                qWarning() << "Variant datatype can't be converted to double";
            }
            break;
        default:
            qWarning()
                    << "unknown datatype: can't write anything to process.";
            break;
    }
}

/****************************************************************************/

/** Increments the current #value and writes it to the process.
 *
 * This does \a not update #value directly.
 */
void ScalarVariant::inc()
{
    Q_D(ScalarVariant);

    writeValue(d->value.toInt() + 1);
}

/****************************************************************************/

bool ScalarVariant::poll()
{
    return QtPdCom::ScalarSubscriber::poll();
}

/****************************************************************************/

/** This virtual method is called by the ProcessVariable, if its value
 * changes.
 */
void ScalarVariant::newValues(std::chrono::nanoseconds ts)
{
    Q_D(ScalarVariant);
    QVariant v;


#ifdef PD_DEBUG_SCALARVARIANT
    uint8_t c;
    d->copyData(c);
    qDebug() << "type " << getVariable().getTypeInfo().type << "value" << c;
#endif

    switch (getVariable().getTypeInfo().type) {
        case PdCom::TypeInfo::boolean_T:
        case PdCom::TypeInfo::uint8_T:
        case PdCom::TypeInfo::uint16_T:
        case PdCom::TypeInfo::uint32_T:
        case PdCom::TypeInfo::uint64_T: {
            uint64_t uInt;
            d->copyData(uInt);
            v = (quint64) (uInt * d->pd_connection.getScale()
                           + d->pd_connection.getOffset());
        } break;
        case PdCom::TypeInfo::int8_T:
        case PdCom::TypeInfo::int16_T:
        case PdCom::TypeInfo::int32_T:
        case PdCom::TypeInfo::int64_T: {
            int64_t sInt;
            d->copyData(sInt);
            // qDebug() << "sint " << sInt;
            v = (qint64) (sInt * d->pd_connection.getScale()
                          + d->pd_connection.getOffset());
        } break;
        case PdCom::TypeInfo::single_T:
        case PdCom::TypeInfo::double_T: {
            double f;
            d->copyData(f);
            v = (double) (f * d->pd_connection.getScale() + d->pd_connection.getOffset());
        } break;
        default:
            qWarning() << "unknown datatype";
            v = (double) 0.0;
            break;
    }

#ifdef PD_DEBUG_SCALARVARIANT
    QString path = getVariable().getPath().data();
    qDebug() << "new value " << v << " for " << path;
#endif

    if (!d->dataPresent) {
        d->dataPresent = true;
#ifdef PD_DEBUG_SCALARVARIANT
        qDebug() << "emit data present: " << d->dataPresent;
#endif
        emit dataPresentChanged(d->dataPresent);
        d->value = v;
        emit valueChanged(d->value);
    }
    else {  // dataPresent == true

        double filterConstant = getFilterConstant();
        // design choice: filtering returns always a double variable
        if (filterConstant > 0.0) {
            v = filterConstant * (v.toDouble() - d->value.toDouble())
                    + d->value.toDouble();
        }

        if (d->value != v) {
            d->value = v;
            emit valueChanged(d->value);
        }
    }

    d->mTime = ts;
#ifdef PD_DEBUG_SCALARVARIANT
    qDebug() << "emit value updated: "
             << std::chrono::duration<double>(mTime).count();
#endif

    emit valueUpdated(getMTimeToDouble());
}

/****************************************************************************/

bool ScalarVariant::getDataPresent() const
{
    const Q_D(ScalarVariant);
    return d->dataPresent;
}

QtPdCom::PdConnection ScalarVariant::getConnection() const
{
    const Q_D(ScalarVariant);
    return d->pd_connection;
}

std::chrono::nanoseconds ScalarVariant::getMTime() const
{
    const Q_D(ScalarVariant);
    return d->mTime;
}

double ScalarVariant::getMTimeToDouble() const
{
    const Q_D(ScalarVariant);
    return std::chrono::duration<double>(d->mTime).count();
}

bool ScalarVariant::hasData() const
{
    const Q_D(ScalarVariant);
    return d->dataPresent;
}

QVariant ScalarVariant::getValue() const
{
    const Q_D(ScalarVariant);
    return d->value;
}
