/*****************************************************************************
 *
 * Copyright (C) 2018-2021  Wilhelm Hagemeister<hm@igh.de>
 *               2018-2022  Florian Pose <fp@igh.de>
 *
 * This file is part of the QtPdCom library.
 *
 * The QtPdCom library is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 *
 * The QtPdCom library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with the QtPdCom Library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 ****************************************************************************/

#ifndef PD_SCALARVARIANT_H
#define PD_SCALARVARIANT_H

#include "ScalarSubscriber.h"
#include "QtPdCom1/Process.h"

#include <QScopedPointer>
#include <QObject>
#include <QVariant>
#include <QQmlEngine>

#include "PdConnection.h"


namespace QtPdCom {

class ScalarVariantPrivate;

/****************************************************************************/

/** Scalar Variant to be used in QML applications
    which is aware of the process to detect connections/or reconnect.

    Import with <tt>import de.igh.qtpdcom 1.4</tt>,
    the QML name of the class is \c PdScalar .
    Only available in Qt6.
 */

class Q_DECL_EXPORT ScalarVariant:
    public QObject,
    public QtPdCom::ScalarSubscriber
{
        Q_OBJECT
        QML_NAMED_ELEMENT(PdScalar)
        QML_ADDED_IN_VERSION(1, 4)

        /** if one want's to update the variable connection values in one call
          * supply a map with
           @code{.json}
            {
                "process": pdProcess, // process instance, if null: use
          default "path": "/osc/cos",   // path (string) "period": 0.2, //
          sample time, 0 for event,
                                      // "poll" for manual poll mode,
                                      // negative values for timer-based
          polling "offset": 0.0,        // offset (double), default 0
                "scale": 1.0,         // scale (double), default 1.0
                "tau":  0.0           // PT1-Filter (double, in seconds) 0: if
          no filter
            }
            @endcode
         */
        Q_PROPERTY(QtPdCom::PdConnection connection READ getConnection WRITE
                           setConnection)

        /* indicates that the process is connected and data is transfered */
        Q_PROPERTY(
                bool connected READ getDataPresent NOTIFY dataPresentChanged)

        Q_PROPERTY(QVariant value READ getValue WRITE setValue NOTIFY
                           valueChanged)
        Q_PROPERTY(double mtime READ getMTimeToDouble NOTIFY valueUpdated)
    public:
        ScalarVariant(QObject *parent = nullptr);
        virtual ~ScalarVariant();

        Q_INVOKABLE bool getDataPresent() const;
        QtPdCom::PdConnection getConnection() const;
        void setConnection(const QtPdCom::PdConnection &);

        void clearData();  // pure-virtual from ScalarSubscriber
        Q_INVOKABLE bool hasData() const;

        QVariant getValue() const;
        Q_INVOKABLE void setValue(QVariant);
        std::chrono::nanoseconds getMTime() const;
        double getMTimeToDouble() const;
        Q_INVOKABLE void inc();
        Q_INVOKABLE bool poll();

    private:
        Q_DECLARE_PRIVATE(ScalarVariant);
        QScopedPointer<ScalarVariantPrivate> const d_ptr;

        Q_DISABLE_COPY(ScalarVariant)

        // pure-virtual from ScalarSubscriber
        void newValues(std::chrono::nanoseconds) override;

    signals:
        void valueChanged(QVariant); /**< Emitted, when the value changes, or
                                       the variable is disconnected. */
        void
        valueUpdated(double); /**< Emitted also when value does not change but
                                got an update from the msr process */
        void pathChanged(QString);
        void dataPresentChanged(bool);
        void connectionUpdated();
};

/****************************************************************************/

}  // namespace QtPdCom


#endif
